/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright © 2002-2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// +-
// |  These functions are overloaded here:
// |    (none, but see mnenhy-headers-msgHdrViewOverlay-loader.js)
// |  Last checked for changes: 2010-02-02
// +-

// +-
// |  the contents of this file will be loaded into the Mnenhy.headers.* context!
// |  ie. foo() is available as goMnenhy.headers.foo()
// +-



// |
// |  constants
// |

// properties
const gkoPropertiesHeaders  = goMnenhy.InitStringBundle("chrome://mnenhy/locale/mnenhy-headers.properties");
const gksFailedNG           = goMnenhy.GetString("NG_failed",    gkoPropertiesHeaders);

// expanded headers from RDF
const gksToggle             = "useToggle";
const gksKnown              = "header:known";
const gksFormat             = "format";
const gksIndex              = "index";
const gksDisplay            = "display";
const gksDefOut             = "text";

// expanded XUL elements
const gksFieldText          = "mail-headerfield";
const gksFieldMail          = "mail-emailheaderfield";
const gksFieldMulti         = "mail-multi-emailHeaderField";
const gksFieldRef           = "mail-messageids-headerfield";
const gksFieldRSS           = "mail-urlfield";  // RSS pseudo field
const gksFieldTags          = "mail-tagfield";  // tags pseudo field
// expanded XUL bindings
const gksClassRef           = "MnenhyRefField";
const gksClassNG            = "MnenhyNGField";
const gksClassText          = "MnenhyTextField";
const gksClassPersist       = "MnenhyToggleState";

// function names
const gksOutputFunction     = "outputFunction";
const gksfOutputRef         = "OutputReferences";
const gksfOutputNews        = "OutputNewsgroups";
const gksfOutputMail        = "OutputEmailAddresses";

// special newsgroup value
const gksPoster             = "poster";



// |
// |  variables
// |

// We need a global window context for most functions, because they
// will call other functions and use objects there. Since we may have
// several windows open, we need a pointer to the current one...
// (see respective calls to SetCurrentWindow)
let goWin = null;

// shortcut to Mnenhy;
let goMnenhy = null;

// robbed dtd strings
let goMnenhyDTDRobbing = null;

// remember where to add our custom header lines
let goInsertHeaderBeforeThis = null;


// |
// |  functions
// |

// set the current window context
function SetCurrentWindow(aoWin)
{
  this.goWin = aoWin || window;
  // set content window for SpaceHit, Mail Start Page etc.
  this.goWin.content = this.goWin.document.getElementById("messagepane").contentWindow;
}


// init
function InitMsgHdrViewOverlay(aoMnenhy)
{
  this.goMnenhy = aoMnenhy;
}


// create an entry for asUniqueName in gExpandedHeaderList
// by reading respective data for asDisplayName from mnenhy.rdf
// if asDisplayName is not given, use asUniqueName instead
function CreateExpandedFromRDF(asUniqueName, asDisplayName)
{
  if (!asDisplayName)
    asDisplayName = asUniqueName;

  // the index names of kofOutput are used in mnenhy.rdf
  const kofOutput = {
                      text: goWin.updateHeaderValue,
                      mail: goWin.OutputEmailAddresses,
                      news: OutputNewsgroups,
                      mid:  OutputReferences,
                      link: OutputLinkify
                    };
  const kbToggle  = !!goMnenhy.m_oRDF.GetItemValue(gksToggle,  [gksKnown, asDisplayName]);
  const ksSort    =   goMnenhy.m_oRDF.GetItemValue(gksIndex,   [gksKnown, asDisplayName]) || asDisplayName;
  const ksOutput  =   goMnenhy.m_oRDF.GetItemValue(gksFormat,  [gksKnown, asDisplayName]);
  const kfOutput  = kofOutput[(ksOutput in kofOutput) ? ksOutput : gksDefOut];

  return {name:          asUniqueName,
          display:       asDisplayName,
          index:         ksSort,
          useToggle:     kbToggle,
          outputFunction:kfOutput};
}


function SortByIndexString(o1,o2)
{
  if (o1.index < o2.index)
    return -1;
  if (o1.index > o2.index)
    return  1;
  return 0;
}


function InitializeHeaderViewTables()
{
  let viewmode = GetHeaderView();
  if (viewmode == gkiViewNoHeaders)
  {
    // delete views
    goWin.gCollapsedHeaderView = {};
    goWin.gExpandedHeaderView  = {};
  }
  else
  {
    // Build HeaderViews
    // Remarks: * if (!g*HeaderView.length) is useless, because g*HeaderView is associative!
    //          * isEmailAddress is used nowhere and hence ignored here

    // TB3 has no gCollapsedHeaderList object anymore
    if ("gCollapsedHeaderList" in goWin)
    {
      // CollapsedHeaderView
      // --> we stick to the list of headers predefined by Mozilla
      for (let index = 0; index < goWin.gCollapsedHeaderList.length; index++)
      {
        goWin.gCollapsedHeaderView[goWin.gCollapsedHeaderList[index].name] =
          new goWin.createHeaderEntry('collapsed', goWin.gCollapsedHeaderList[index]);
      }
    }

    // ExpandedHeaderView
    // --> we prepare some headers in advance and sort them
    goWin.gExpandedHeaderList = []; // delete Mozilla's definitions

    let rsHeaderList = "";
    if (viewmode >= gkiViewCustomBase)
    {
      rsHeaderList = goMnenhy.m_oRDF.GetNodeList("header:custom" + (viewmode - gkiViewCustomBase));
    }
    else
    {
      // default list of headers to show, e.g. in Normal mode
      rsHeaderList = ["subject", "from", "reply-to", "content-base", "date", "to", "cc", "bcc", "newsgroups", "followup-to", "tags"];
      // we directly read the prefs here, since only SM has global variables for those
      if (goMnenhy.GetPref("mailnews.headers.showOrganization", goMnenhy.kiPrefBool))
        rsHeaderList.push("organization");
      if (goMnenhy.GetPref("mailnews.headers.showUserAgent", goMnenhy.kiPrefBool))
        rsHeaderList.push("user-agent");
      if (goMnenhy.GetPref("mailnews.headers.showMessageId", goMnenhy.kiPrefBool))
        rsHeaderList.push("message-id");
      if (goMnenhy.GetPref("mailnews.headers.showReferences", goMnenhy.kiPrefBool))
        rsHeaderList.push("references");
      if (goMnenhy.GetPref("mailnews.headers.showSender", goMnenhy.kiPrefBool))
        rsHeaderList.push("sender");
    }
    for (let iHeader in rsHeaderList)
      goWin.gExpandedHeaderList[iHeader] = CreateExpandedFromRDF(rsHeaderList[iHeader])
    goWin.gExpandedHeaderList.sort(SortByIndexString)
    // sync current gExpandedHeaderView and gExpandedHeaderList
    for (let index = 0; index < goWin.gExpandedHeaderList.length; index++)
    {
      const ksHeaderName = goWin.gExpandedHeaderList[index].name;
      goWin.gExpandedHeaderView[ksHeaderName] = new CreateExpandedView(goWin.gExpandedHeaderList[index]);
    }
  } // Build HeaderViews
}


function OutputLinkify(aoHeaderEntry, aoHeaderValue)
{
  // linkification is done via class, no special needs here
  goWin.updateHeaderValue(aoHeaderEntry, aoHeaderValue);
}


function OutputReferences(aoHeaderEntry, aoHeaderValue)
{
  // we replaced the MID binding to persist the twisty state,
  // hence we need have our own output function here as well
  
  // do it like NS 4.x did: show linkified numbers instead of actual MID
  // <MID> => [1] -> news://<currentServer>/MID
  // this Link should be copyable and clicking should search the MID
  aoHeaderEntry.enclosingBox.clearHeaderValues();
  let rsRefs = aoHeaderValue.match(/[^<>\s]+/g);  // MIDs without <>
  for (let iRef = 0; iRef < rsRefs.length; ++iRef)
  {
    let oRef = {
                sMID:   rsRefs[iRef],
                iIndex: iRef + 1
               };
    aoHeaderEntry.enclosingBox.addAddressView(oRef);
  }
  aoHeaderEntry.enclosingBox.buildViews(-1);  // show all
}


function OutputNewsgroups(aoHeaderEntry, aoHeaderValue)
{
  // do it like NS 4.x did: show linkified newsgroups
  // <NG> => [NG] -> news://<currentServer>/NG
  // this Link should be copyable and clicking should call the NG, if present
  aoHeaderEntry.enclosingBox.clearHeaderValues();
  let rsNGs = aoHeaderValue.split(",");
  for (let iNG = 0; iNG < rsNGs.length; ++iNG)
  {
    aoHeaderEntry.enclosingBox.addAddressView(rsNGs[iNG]);
  }
  aoHeaderEntry.enclosingBox.buildViews(-1);  // show all
}


function UpdateMessageHeaders()
{
  // reset x-face overrule of AIM buddy icons
  goMnenhy.Log("UpdateMessageHeaders")
  if ("gfSetFromBuddyIcon" in goWin)
  {
    goWin.setFromBuddyIcon = goWin.gfSetFromBuddyIcon;
    delete goWin.gfSetFromBuddyIcon;
  }
  // TB3 has killed the AIM buddy icon handling, so we may need to hide it here
  // (This has to happen before we call oHeaderEntry.outputFunction for the from: header below!)
  if (!("x-face" in goWin.currentHeaderData))
  {
    const koAIMBuddyIcon = goWin.document.getElementById("fromBuddyIcon");
    if (koAIMBuddyIcon)
      koAIMBuddyIcon.setAttribute("src", "");
  }

  // TB3 has no collapsed headers anymore, thus we need to check the 
  // existence of the variable before using it
  let bCollapsedHeaderViewMode = ("gCollapsedHeaderViewMode" in goWin) 
                               && goWin.gCollapsedHeaderViewMode;

  // iterate over each header we received and see if we
  // have a matching entry in each header view table...
  for (let sHeaderName in goWin.currentHeaderData)
  {
    goMnenhy.Log("sHeaderName="+sHeaderName)
    let oCurrentHeader = goWin.currentHeaderData[sHeaderName];
    let oHeaderEntry   = null;
    // some headers need special attention
    switch (sHeaderName)
    {
      case "subject":
      {
        // show subject in window title bar
        try
        {
          if (goWin.gDBView.keyForFirstSelectedMessage == goWin.nsMsgKey_None)
          {
            let oFolder = null;
            if (goWin.gCurrentFolderUri)
              oFolder = goWin.RDF.GetResource(goWin.gCurrentFolderUri).QueryInterface(Components.interfaces.nsIMsgFolder);
            goWin.setTitleFromFolder(oFolder, oCurrentHeader.headerValue);
          }
        }
        catch(ignored){}
        break;
      }

      case "x-face":
      {
        if ("xface" in goMnenhy.headers)
        {
          // set x-face
          let oFromBuddyIcon = goWin.document.getElementById("fromBuddyIcon");
          const koComputedStyle = goWin.getComputedStyle(oFromBuddyIcon, null);
          const ksFaceURL = goMnenhy.headers.xface.FaceURL(oCurrentHeader.headerValue, koComputedStyle);
          oFromBuddyIcon.setAttribute("src", ksFaceURL);
          // overrule AOL buddy icons (unneeded for TB3)
          goWin.gfSetFromBuddyIcon = ("setFromBuddyIcon" in goWin) && goWin.setFromBuddyIcon;
          goWin.setFromBuddyIcon = function(ignored){};
        }
      }
    }

    if (bCollapsedHeaderViewMode && !goWin.gBuiltCollapsedView)
    {
      if (sHeaderName in goWin.gCollapsedHeaderView)
      {
        // use existing view
        oHeaderEntry = goWin.gCollapsedHeaderView[sHeaderName];
      }
    }
    else if (!bCollapsedHeaderViewMode && !goWin.gBuiltExpandedView)
    {
      if (sHeaderName in goWin.gExpandedHeaderView)
      {
        // use existing view
        oHeaderEntry = goWin.gExpandedHeaderView[sHeaderName];
      }
      else
      {
        // non-precreated headers are only shown if the view mode is gkiViewAllHeaders
        // or if the 'real' name is valid for display
        const kiViewMode = GetHeaderView();
        if (    (kiViewMode == gkiViewAllHeaders)
             || (oCurrentHeader.headerName in goWin.gExpandedHeaderView))
        {
          // create new view
          const koExpandedListItem = CreateExpandedFromRDF(sHeaderName, oCurrentHeader.headerName);
          oHeaderEntry = new CreateExpandedView(koExpandedListItem)
          goWin.gExpandedHeaderView[sHeaderName] = oHeaderEntry;
        }
      } // create new view
    } // if we are in expanded view....

    if (oHeaderEntry)
    {
      // to allow for custom header styles, add headername and value as attribute
      // e.g.
      //      [header="subject"][headervalue*="Mozilla"]
      //      {
      //        color:yellow!important;
      //      }
      // In TB3 enclosingBox is inside enclosingRow, in SM these are the same
      oHeaderEntry.enclosingBox.setAttribute("header",      sHeaderName);
      oHeaderEntry.enclosingBox.setAttribute("headervalue", oCurrentHeader.headerValue);
      if (oHeaderEntry.enclosingRow)
      {
        oHeaderEntry.enclosingRow.setAttribute("header",      sHeaderName);
        oHeaderEntry.enclosingRow.setAttribute("headervalue", oCurrentHeader.headerValue);
      }
      oHeaderEntry.outputFunction(oHeaderEntry, oCurrentHeader.headerValue);
      oHeaderEntry.valid = true;
    }
  } // for (let sHeaderName in currentHeaderData)

  if (bCollapsedHeaderViewMode)
    goWin.gBuiltCollapsedView = true;
  else
    goWin.gBuiltExpandedView = true;

  // now update the view to make sure the right elements are visible
  if (goMnenhy.installed.IsThunderbird())
    goWin.updateExpandedView();
  else
    goWin.updateHeaderViews();
}


// helper function to find the correct target parent node for certain headers
function GetTargetParentNodeForHeader(abIsFirstHeader)
{
  // default for SM2
  let sTargetID = "expandedHeaders";
  if (goMnenhy.installed.IsThunderbird())
  {
    if (abIsFirstHeader)
      sTargetID = "expandedHeaderRows";
    else
      sTargetID = "expandedHeader2Rows";
  }
  return goWin.document.getElementById(sTargetID);
}


// crudely hide the headerpane
// returns if we're in no header mode
function ForceHideHeaderPane()
{
  let oElem = goWin.document.getElementById("msgHeaderView");
  let rv = (GetHeaderView() == gkiViewNoHeaders);
  if (rv)
  {
    // <sing>no headers, no box...</sing> ;-)
    oElem.setAttribute("collapsed","true");
  }
  else
  {
    oElem.removeAttribute("collapsed");
  }
  return rv;
}


// modify the messageHeaderSink's onStartHeaders function
// if we don't, the header view does not get updated when we change from all to custom
function OnStartHeaders()
{
  goMnenhy.Log("OnStartHeaders")

  if ("mSaveHdr" in goWin.messageHeaderSink)  // XXX TB 1.0 hack
  {
    goWin.messageHeaderSink.mSaveHdr = null;  // that's how it should be
    goWin.mSaveHdr = null;                    // that's what they did :(
  }

  if (!ForceHideHeaderPane())
  {
    // clear out any pending collected address timers...
    if (("gCollectAddressTimer" in goWin) && goWin.gCollectAddressTimer)
    {
      goWin.gCollectAddress = "";
      goWin.clearTimeout(goWin.gCollectAddressTimer);
      goWin.gCollectAddressTimer = null;
    }

    let iHeaderView = GetHeaderView();
    if (iHeaderView != goWin.giMnenhyHeaderView)
    {
      // kill all widget headers to enforce rebuild
      goWin.hideHeaderView(goWin.gExpandedHeaderView);
      let oTopViewNode = null;
      if (goMnenhy.installed.IsMozilla())
      {
        // SeaMonkey
        oTopViewNode = GetTargetParentNodeForHeader(true);
        // there may be non-widget elements in the list
        // we push them just to the end
        let oChild = oTopViewNode.lastChild, oNextChild;
        while (oChild)
        {
          oNextChild = oChild.previousSibling;
          // only kill known widget types
          switch (oChild.localName)
          {
            case gksFieldText:
            case gksFieldMail:
            case gksFieldRef:   // messageids
            case gksFieldMulti:
            case gksFieldRSS:   // RSS pseudo header
            case gksFieldTags:  // tags pseudo header
              oTopViewNode.removeChild(oChild);
              break;
            default:
          }
          oChild = oNextChild;
        }
      }
      else if (goMnenhy.installed.IsThunderbird())
      {
        // completely clear both grids
        oTopViewNode = GetTargetParentNodeForHeader(true);
        while (oTopViewNode.lastChild)
          oTopViewNode.removeChild(oTopViewNode.lastChild);
        if (!oTopViewNode.hasAttribute("pack"))
        {
          // correct packing if necessary
          oTopViewNode.setAttribute("pack", "end");
        }
        oTopViewNode = GetTargetParentNodeForHeader(false);
        while (oTopViewNode.lastChild)
          oTopViewNode.removeChild(oTopViewNode.lastChild);
        // hide the date box, we won't use it
        goWin.document.getElementById("dateLabel").hidden = true;
      }
      goInsertHeaderBeforeThis = oTopViewNode && oTopViewNode.firstChild;
      goWin.gExpandedHeaderView = {};
      goWin.initializeHeaderViewTables();
      goWin.giMnenhyHeaderView = iHeaderView;
    }

    // set gViewAllHeaders (for backwards and Enigmail compatibility)
    goWin.gViewAllHeaders = true;

    goWin.ClearCurrentHeaders();
    goWin.gBuiltExpandedView = false;
    goWin.gBuiltCollapsedView = false;
    goWin.gBuildAttachmentsForCurrentMsg = false;
    goWin.gBuildAttachmentPopupForCurrentMsg = true;
    goWin.ClearAttachmentList();
    goWin.ClearEditMessageBox();
    goWin.gMessageNotificationBar.clearMsgNotifications();

    for (let index in goWin.gMessageListeners)
      goWin.gMessageListeners[index].onStartHeaders();
  }
}


// modify the messageHeaderSink's processHeaders function
function ProcessHeadersNew(arHeaderNames, arHeaderValues, abDontCollectAddress)
{
  goMnenhy.Log("ProcessHeadersNew")
  goWin.messageHeaderSink.onStartHeaders();
  // process each header
  while (arHeaderNames.hasMore())
  {
    let oHeader = {headerName:  arHeaderNames .getNext(),
                   headerValue: arHeaderValues.getNext()};
    if (oHeader.headerName)
      ProcessHeader(oHeader, abDontCollectAddress);
  }
  // process message tags as if they were headers in the message
  if ("SetTagHeader" in goWin)
    goWin.SetTagHeader();
  goWin.messageHeaderSink.onEndHeaders();

  if (goMnenhy.installed.IsThunderbird())
  {
    // Really, we want to hide the header pane, believe it!
    ForceHideHeaderPane();
  }
}


// if there're multiple headers of the same kind, we usually show them all.
// but in case of to, cc and bcc, we merge
// !!! the bug fixed here is fixed in the Mozilla trunk since 2002-12-10 (->1.3a) !
function ProcessHeader(aoHeader, abDontCollectAddress)
{
  // for consistancy sake, let's force all header names to be lower case so
  // we don't have to worry about looking for: Cc and CC, etc.
  let sLowerCaseHeaderName = aoHeader.headerName.toLowerCase();
  aoHeader.headerName = sLowerCaseHeaderName;
  let sUniqueName = sLowerCaseHeaderName;
  let bOldHeader = (sUniqueName in goWin.currentHeaderData);
  if (bOldHeader)
  {
    // need to create an unique name
    // special handling for certain headers
    switch (sLowerCaseHeaderName)
    {
      case "to":
      case "cc":
      case "bcc":
          goWin.currentHeaderData[sLowerCaseHeaderName].headerValue = goWin.currentHeaderData[sLowerCaseHeaderName].headerValue
                                                                    + ',' + aoHeader.headerValue;
          break;

      default:
      {
        let iUID = 0;
        do
        {
          sUniqueName = sLowerCaseHeaderName + iUID;
          ++iUID;
        }
        while (sUniqueName in goWin.currentHeaderData);
        bOldHeader = false; // create this new header
      }
      break;
    }
  }

  if (!bOldHeader)
  {
    goWin.currentHeaderData[sUniqueName] = aoHeader;
  }

  if (sLowerCaseHeaderName == "from")
  {
    // TB3 doesn't collect these addresses
    if (!goMnenhy.installed.IsThunderbird() && aoHeader.headerValue)
    {
      try
      {
        if (!abDontCollectAddress && (goWin.gCollectIncoming || goWin.gCollectNewsgroup || goWin.gCollectOutgoing))
        {
          if (!goWin.abAddressCollector) // global let
            goWin.abAddressCollector = Components.classes["@mozilla.org/addressbook/services/addressCollector;1"]
                                                 .getService(Components.interfaces.nsIAbAddressCollector);
          // collect, add card if doesn't exist and gCollectOutgoing is set,
          // otherwise only update existing cards, unknown preferred send format
          goWin.gCollectAddress = aoHeader.headerValue;
          goWin.gCollectAddressTimer = goWin.setTimeout(CallAddressCollector,
                                                        2000,
                                                        goWin.gCollectAddress,
                                                        !goWin.gCollectOutgoing);
        }
      }
      catch (ignored){}
    }
  }
}


function CallAddressCollector(asAddress, abCreateCard)
{
  // make sanity checks because the window maybe closed before this is called
  if (("abAddressCollector" in goWin) && goWin.abAddressCollector && ("collectUnicodeAddress" in goWin.abAddressCollector))
  {
    // since about 1.8a, collectUnicodeAddress takes a third parameter, but superfluous
    // parameters do not matter in JS, so this should be backwards compatible
    goWin.abAddressCollector
         .collectUnicodeAddress(asAddress,
                                abCreateCard,
                                Components.interfaces.nsIAbPreferMailFormat.unknown);
  }
}


function OnPopupShowingText(event)
{
  let oMenuPopUp = event.originalTarget;
  if (oMenuPopUp.id == "MnenhyTextPopup")
  {
    // disable copy selection if there is no selection
    let oSel = oMenuPopUp.ownerDocument.defaultView.getSelection();
    oMenuPopUp.firstChild.disabled = !oSel || oSel.isCollapsed;

    // if we haven't done so yet, drag in some other menuitems
    // we need to do this here, since overlaying doesn't work
    // (content can't exist twice when overlaying)
    if (!("mnenhyMIDSAdded" in oMenuPopUp) && goWin)
    {
      // message-id search
      let oMIDS = goWin.document.getElementById("mnenhyMIDContextMenu");
      if (oMIDS)
      {
        oMenuPopUp.appendChild(goWin.document.createElement("menuseparator"));
        oMenuPopUp.appendChild(oMIDS.cloneNode(true));
        oMenuPopUp.mnenhyMIDSAdded = true;
      }
    }
    if (!("mnenhyTextCodecsAdded" in oMenuPopUp) && goWin)
    {
      // text codecs
      let oEnc = goWin.document.getElementById("mnenhyCodecsEncMenu");
      let oDec = goWin.document.getElementById("mnenhyCodecsDecMenu");
      if (oEnc && oDec)
      {
        oMenuPopUp.appendChild(goWin.document.createElement("menuseparator"));
        oMenuPopUp.appendChild(oEnc.cloneNode(true));
        oMenuPopUp.appendChild(oDec.cloneNode(true));
        oMenuPopUp.mnenhyTextCodecsAdded = true;
      }
      
    }
  }
}


function CreateExpandedView(oHeaderEntry)
{
  if (!oHeaderEntry)
    return;

  let sHeaderName    = oHeaderEntry.name;
  let sPartialIDName = "expanded" + sHeaderName;
  this.enclosingBox  = goWin.document.getElementById(sPartialIDName + "Box");
  this.textNode      = goWin.document.getElementById(sPartialIDName + "Value");
  this.isValid       = false;

  // if we already have a view box, add our class,
  // else create a new box
  if (this.enclosingBox)
  {
    // class override
    AddExpandedBoxFeatures(this.enclosingBox, oHeaderEntry)
  }
  else
  {
    // new box
    this.enclosingBox = new CreateExpandedBox(oHeaderEntry);
  }

  // TB3 needs an enclosingRow as well
  this.enclosingRow = goWin.document.getElementById(sPartialIDName + "Row");
  if (!this.enclosingRow)
    this.enclosingRow = this.enclosingBox;  // allow Mnenhy to just use row

  if ("useToggle" in oHeaderEntry)
  {
    this.useToggle = oHeaderEntry.useToggle;
    // find the toggle icon in the document
    if (this.useToggle)
    {
      this.textNode = this.enclosingBox.emailAddresses;
    }
  }
  else
   this.useToggle = false;

  if ("useShortView" in oHeaderEntry)
  {
    this.useShortView = oHeaderEntry.useShortView;
    this.enclosingBox.emailAddressNode = this.textNode;
  }
  else
    this.useShortView = false;

  if (gksOutputFunction in oHeaderEntry)
    this.outputFunction = oHeaderEntry.outputFunction;
  else
    this.outputFunction = updateHeaderValue;
}


function CreateExpandedBox(oHeaderEntry)
{
  if (!oHeaderEntry)
    return null;

  // create element according to outputFunction
  let sOutputFunction = "";   // use default output
  if ((gksOutputFunction in oHeaderEntry) && (oHeaderEntry.outputFunction))
    sOutputFunction = oHeaderEntry.outputFunction.name;

  let sNewElement;
  switch (sOutputFunction)
  {
    case gksfOutputRef:
      sNewElement = gksFieldMulti;
      break;

    case gksfOutputNews:
      sNewElement = gksFieldMulti;
      break;

    case gksfOutputMail:
      if ((!("useToggle" in oHeaderEntry)) || (!oHeaderEntry.useToggle))
        sNewElement = gksFieldMail;  // one address
      else
        sNewElement = gksFieldMulti; // multiple
      break;

    default:
      // specialcase the tags pseudo header
      if (oHeaderEntry.name == "tags")
        sNewElement = gksFieldTags;
      else
        sNewElement = gksFieldText;
  }

  // create new view and set attributes
  let sDisplaySuffix = goMnenhy.installed.IsThunderbird() ? " " : ": ";
  // 1. check mnenhy.rdf for user-defined header display names
  let sDisplayName = goMnenhy.m_oRDF.GetItemValue(gksDisplay, [gksKnown, oHeaderEntry.display]);
  if (sDisplayName)
  {
    sDisplayName += sDisplaySuffix;
  }
  // 2. else look for Mozilla-defined localized header names
  else
  {
    if (!goMnenhyDTDRobbing)
      goMnenhyDTDRobbing = goWin.document.getElementById("mnenhydtdrobbing");
    if (goMnenhyDTDRobbing)
    {
      const ksHeaderAttr = oHeaderEntry.name.replace("-", "");
      sDisplayName = goMnenhyDTDRobbing.getAttribute(ksHeaderAttr);
    }
  }
  // 3. else create our own nice display
  if (!sDisplayName)
  {
    if (goMnenhy.installed.IsThunderbird())
      sDisplayName = oHeaderEntry.display.toLowerCase() + sDisplaySuffix;
    else
      sDisplayName = goMnenhy.Capitalize(oHeaderEntry.display) + sDisplaySuffix;
  }

  const ksBoxIdName   = "expanded" + oHeaderEntry.name + "Box"; 
  let oNewView = goWin.document.createElement(sNewElement);
  oNewView.setAttribute("id",    ksBoxIdName);
  oNewView.setAttribute("label", sDisplayName);
  oNewView.setAttribute("keywordrelated", true);
  oNewView.collapsed = true;

  AddExpandedBoxFeatures(oNewView, oHeaderEntry);

  // insert this new element at the correct position into the view...
  let oElementToInsert = oNewView;
  let oInsertHeaderBeforeThis = goInsertHeaderBeforeThis;
  let oTopViewNode = GetTargetParentNodeForHeader(true);
  if (goMnenhy.installed.IsThunderbird())
  {
    // If the slot for the first header is still free, use it, else pass it on
    // to the main grid.
    if (!oTopViewNode.firstChild)
    {
      // Don't use goInsertHeaderBeforeThis, it always relates to the main grid!
      goInsertHeaderBeforeThis = null;
    }
    else
    {
      // pass the element on to the main grid
      oTopViewNode = GetTargetParentNodeForHeader(false);
    }

    // TB3 shows its headers in a grid, with the header name column split off.
    // Hence we need to create row and header name column here as well.
    oElementToInsert = goWin.document.createElement("row");
    oElementToInsert.setAttribute("id", "expanded" + oHeaderEntry.name + "Row");
    let oNameCol = goWin.document.createElement("hbox");
    oNameCol.setAttribute("id", "expanded" + oHeaderEntry.name + "Label");
    oNameCol.setAttribute("label", sDisplayName);
    oNameCol.setAttribute("pack",  "end");
    oNameCol.setAttribute("class", "MnenhyToggleState");
    oElementToInsert.appendChild(oNameCol);
    oElementToInsert.appendChild(oNewView);
    // TB3 hides unfitting grid rows, not the actual elements
    oNewView.collapsed = false;
    //oNewView.setAttribute("align", "center");
  }
  // If the oInsertHeaderBeforeThis exists, but does not belong to the correct
  // oTopViewNode, we let it point to the beginning of the current child list.
  // That way, any non-deleted remainder elements will end up at the end.
  if (oInsertHeaderBeforeThis && oInsertHeaderBeforeThis.parentNode != oTopViewNode)
    oInsertHeaderBeforeThis = oTopViewNode.firstChild;
  oTopViewNode.insertBefore(oElementToInsert, oInsertHeaderBeforeThis);
  if (!goInsertHeaderBeforeThis)
   goInsertHeaderBeforeThis = oNewView.nextSibling;
  return oNewView;
}


// set new features like class, context, etc. according to the output function
function AddExpandedBoxFeatures(oBox, oHeaderEntry)
{
  if (oBox && oHeaderEntry)
  {
    let sOutputFunction = "";   // use default output
    if ((gksOutputFunction in oHeaderEntry) && (oHeaderEntry.outputFunction))
      sOutputFunction = oHeaderEntry.outputFunction.name;

    // add new class
    let sFunctionClass;
    switch (sOutputFunction)
    {
      case gksfOutputRef:
        sFunctionClass = gksClassRef;
        break;
      case gksfOutputNews:
        sFunctionClass = gksClassNG;
        break;
      case gksfOutputMail:
        sFunctionClass = gksClassPersist;
        break;
      default:
        if (oHeaderEntry.name == "tags")
          sFunctionClass = "";
        else
          sFunctionClass = gksClassText;
    }

    // add class
    let sNewClass = oBox.getAttribute("class");
    if (sFunctionClass && (String(sNewClass).indexOf(sFunctionClass) < 0))
    {
      sNewClass += " " + sFunctionClass;
      sNewClass = sNewClass.replace(/^\s+|\s+$/g, "");      // trim whitespace
      oBox.setAttribute("class", sNewClass);
    }

    // after work party ;-)
    if (sFunctionClass == gksClassText)
    {
      // now that we ain't partying in the switch default branch anymore,
      // keep the following code commented out for future reference only
      //      // Since we're constructing the oBox before it is shown, the XBL properties
      //      // aren't in place yet and assignment to them creates object members!
      //      // You can't |delete| properties, so it's safe to |delete| this members...
      //      // (I *really* do *hate* such bugs! :( )
      //      delete oBox.doLinkify;
      oBox.doLinkify = (sOutputFunction == "OutputLinkify");
    }

    // if we have localized header abbreviations (like "An: Sie" vs. "Von: Ihnen"),
    // then we need the headerName as context
    oBox.headerName = oHeaderEntry.name;
  }
}


function OnMnenhyRefClick(event)
{
  // see also msgHdrViewOverlay.js:MessageIdClick
  if (!event.button)
  {
    // left click
    let sMID = event.view.GetMessageIdFromNode(event.explicitOriginalTarget, true);
    event.view.OpenMessageForMessageId(sMID);
  }
}


// try to select the folder given by asURI
function MnenhySelectFolder(asURI)
{
  let bSuccess = false;
  if (goMnenhy.installed.IsThunderbird())
  {
    // TB3 doesn't raise an exception for selecting non-existing folders
    bSuccess = goWin.gFolderTreeView.selectFolder(goWin.GetMsgFolderFromUri(asURI));
  }
  else
  {
    try
    {
      goWin.SelectFolder(asURI);
      bSuccess = true;
    }
    catch(ex) {}
  }
  return bSuccess;
}


function OnMnenhyNGClick(event)
{
  if (!event.button)
  {
    let sNG = event.explicitOriginalTarget.getAttribute("sNG");
    if (sNG == gksPoster)
    {
      // if clicked NG is poster, reply to sender
      goWin.goDoCommand('cmd_replySender');
    }
    else
    {
      goMnenhy.SetWaitCursor(goWin);

      let oServer  = goWin.gDBView.msgFolder.server;
      let sURI     = oServer.serverURI + "/" + sNG;
      let bSuccess = MnenhySelectFolder(sURI);
      if (!bSuccess)
      {
        // could not select folder, maybe 'cos it ain't there yet
        goMnenhy.UnsetWaitCursor(goWin);
        if (sURI.substr(0,7) == "news://")
        {
          // only try if in news folder
          let oSubServ = oServer.QueryInterface(Components.interfaces.nsISubscribableServer);
          if (GetSubscribeConfirmation(sNG))
          {
            goMnenhy.SetWaitCursor(goWin);
            try
            {
              oSubServ.subscribe(sNG);
              oSubServ.commitSubscribeChanges();
              bSuccess = MnenhySelectFolder(sURI); // retry selecting folder
            }
            catch(ex) {}
            if (!bSuccess)
            {
              goMnenhy.UnsetWaitCursor(goWin);
              goMnenhy.AlertService(gksFailedNG.replace("%ng", sNG), null, goWin);
            } // no group found on server
          } // subscribing allowed
        } // we are on a news server
      }
      goMnenhy.UnsetWaitCursor(goWin);
    }
  }
}


function GetSubscribeConfirmation(sNG)
{
  try
  {
    const koNewsBundle = goMnenhy.InitStringBundle("chrome://messenger/locale/news.properties");
    let s = koNewsBundle.GetStringFromName("autoSubscribeText").replace("%1$S", sNG);
    return goMnenhy.ConfirmService(s, null, goWin);
  }
  catch(error)
  {
    return false;
  }
}


// CopyNG to the clipboard
function CopyNG(aoNG, abAll)
{
  if (aoNG)
  {
    if (goMnenhy.installed.IsThunderbird())
      aoNG = aoNG.parentNode.parentNode;
    let sNG = "";
    if (abAll)
    {
      let roNGs = aoNG.parentNode.parentNode.mAddresses;
      for (let i = 0; i < roNGs.length; ++i)
        sNG += roNGs[i] + ",";
      sNG = sNG.replace(/,+$/, "");
    }
    else
    {
      sNG = aoNG.getAttribute("sNG");
    }
    goMnenhy.CopyStringToClipboard(sNG);
  }
}


// CopyText to the clipboard
function CopyText(aoText, abAll)
{
  if (aoText)
  {
    let sText = "";
    if (!abAll)
    {
      // if selection requested and non-empty, copy it (text only!)
      const oSel = aoText.ownerDocument.defaultView.getSelection();
      if (oSel && !oSel.isCollapsed)
        sText = oSel.toString();
    }
    if (!sText)
    {
      // if no selection requested or empty, copy original header value
      sText = aoText.ownerDocument.getBindingParent(aoText).headerValue;
    }
    goMnenhy.CopyStringToClipboard(sText);
  }
}
